# -*- coding: cp1252 -*-

# Integrate double layer
#   dx/d(psi)= (2RT/epw * SUM c_i*[exp(-z_i F psi / RT) - 1])^-.5;

from numpy import exp, linspace, concatenate, zeros, arange;
from pylab import *;
from scipy.integrate import solve_ivp;
from scipy.interpolate import InterpolatedUnivariateSpline;
from scipy.optimize import minimize_scalar;

#     J/kmol         C/kmol          e_w    e_0
RT = 2477.57e3; F = 96485.e3; epw = 78.5 * 8.854e-12;

# closest distance to surface = Debye-Huckel a ...
dha_Na = 4.0e-10;  dha_K = 3.501e-10; dha_Mg = 5.5e-10; dha_Ca = 4.5e-10; dha_Cl = 3.5e-10;

# conc's in kmol/m3 at x = inf, psi = 0 ...
Na_0 = 0.0037; K_0 = 0.0003; Mg_0 = 0.0005;  Ca_0 = 0.0025;  Cl_0 = 0.01;
psi0 = -0.09137; 

#  m: symbol = 0     c_0=1    dha=2     charge=3    int dl=4    beta=5     Ki\Na=6
m = [['Na', 'K', 'Mg', 'Ca', 'Cl'], [Na_0, K_0, Mg_0, Ca_0, Cl_0], [dha_Na, dha_K, dha_Mg, dha_Ca, dha_Cl], [1, 1, 2, 2, -1], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0]];
anion = 4;

I = 0;
for i in range (len(m[0])):
    if i < len(m[0]):
        if m[1][i] == 0:
            for i1 in range(len(m)): del m[i1][i];
            i -= 1;
        else: I += m[1][i] * m[3][i]**2;
        #
        if m[3][i] < 0: anion = i;
#
I /= 2; t1 = 0.1 / (0.117 * I**0.5)
# Gouy eqn for monovalent cations in dl:
# psi0 = -RT / F * log(t1 + (t1**2 + 1)**0.5); 

dha = m[2].copy();
dha.sort();
min1 = dha[0];
max1 = max(dha);

n = len(m[0]);
def inv_dl(t, y): # dx/d(psi) = (2RT/epw * SUM c_i*[exp(-z_i F psi / RT) - 1])^-.5
    psi = t; x = y;
    d = exp(-F*psi/RT); d1 = 0;
    for i in range(n):
        if x >= m[2][i]: d1 += m[1][i] * (d**(m[3][i]) - 1)
    if d1 == 0: d1 = 1e-21;
    return (2 * RT / epw * abs(d1))**(-0.5);
#

def f(psi0 = -0.05, tsp = 1):
    psi3 = -1e-6;
    dx = 1e-12;
    dha = m[2].copy();
    min1 = min(dha);
    ## When called from minimize, args are in list/ndarray, extract psi0;
    if not isinstance(psi0, float):
        if len(psi0) > 1: tsp = psi0[1];
        #
        psi0 = psi0[0]; 
    if tsp != None:
        ## for detailed plotting, use option: t_eval = tsp.
        psi1 = psi0 / 1.01; psi2 = psi0 / 1.6; nf = 500; nf1 = 1
        tsp = arange(psi0, psi1, (-psi0 + psi1) / nf); nf /= nf1; nf1 = 1
        tsp = concatenate([tsp, linspace(psi1 * 0.99, psi2, int(nf))]); nf /= nf1;
        tsp = concatenate([tsp, linspace(psi2 * 0.99, psi3, int(nf))]);
    #
    # print(psi0)
    s = solve_ivp(lambda t, y: inv_dl(t, y), [psi0, psi3], [min1 - 1e-16], method = 'RK45', t_eval = tsp, first_step = dx, atol = 1e-16, rtol = 1e-6)
    global x, psi, c, c1, CEC, CEC1;
    x = s.y[0]; psi = s.t;
    d = exp(-F * psi/RT); n1 = len(d);
    c = zeros((n, n1));
    c1 = zeros((n, n1));
    denom = zeros(n1);
    for i in range(n):
        for i1 in range(n1): 
            if x[i1] < m[2][i]: c[i][i1] = 0;
            else: c[i][i1] = m[1][i] * d[i1]**m[3][i];
            if m[3][i] > 0: denom[i1] += m[3][i] * c[i][i1];
    #
    # subtract the anion charge from the cations, distributed by equivalents: 
    # cNa = Na_0*d - Na_0*d / (Na_0*d + 2Ca_0*d^2) * Cl_0/d
    # cCa = Ca_0*d^2 - Ca_0*d^2 / (Na_0*d + 2Ca_0*d^2) * Cl_0/d
    # Cl is the last of n
    for i in range(n - 1):
        for i1 in range(n1): 
            a = max(m[2][i], m[2][anion])
            if c[i][i1] > 0:
                if x[i1] < m[2][anion]: c1[i][i1] = c[i][i1];
                else:
                    c1[i][i1] = c[i][i1] - c[i][i1] / denom[i1] * c[anion][i1];
                    if c1[i][i1] < 0: c1[i][i1] = 0;
    #
    CEC = 0; CEC1 = 0
    for i in range(n):
        m[4][i] = InterpolatedUnivariateSpline(x, c1[i], k = 1).integral(0, 1);
        m[5][i] = m[4][i] * m[3][i];
        if m[3][i] > 0: CEC += m[5][i];
        
        # CEC1 += (m[3][i] * InterpolatedUnivariateSpline(x, c[i], k = 1).integral(0, 1));
    #
    # print('CEC = ', format(CEC1 * 1e3, '0.2e'), ' eq/m2')
    # print('Anion corrected CEC = ', format(CEC * 1e3, '0.2e'), ' eq/m2')
    return (CEC * 1e9 - 1)**2 # 1e-6 eq/m2 is target for minimize
#

from time import time
st = time()
psi0 = -0.09
f(psi0);

# r = minimize_scalar(f, bounds = (-0.15, -0.05), method = 'bounded', tol = 1e-7);
# print(r)
print(' Time to calculate the dl:', format(time() -st, '0.2e'), ' s')
x *= 1e9;
figure(1)
plot(x, psi);
xlabel('Distance from charged surface / nm');
ylabel('Potential / V');
xlim(0, 20)

figure(2)
for i in range(n):
    f = 1; ls = ''
    if m[0][i] == 'K': f = 0.2; ls = ' * 5';
    if m[0][i] == 'Ca': f = 5; ls = ' / 5';
    if m[0][i] == 'Cl': f = 1/30; ls = ' * 30';
    plot(x, c[i] / f, label = m[0][i] + ls)
xlabel('Distance from charged surface / nm');
ylabel('mol/L');
xlim(0, 2); #ylim(0, 0.5)
legend()

print(' Number of points in c-x:', format(len(x), '0.0f'));

show()
